// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

package com.starrocks.sql.optimizer.task;

import com.google.common.collect.Lists;
import com.starrocks.common.Pair;
import com.starrocks.sql.optimizer.GroupExpression;
import com.starrocks.sql.optimizer.OptExpression;
import com.starrocks.sql.optimizer.operator.pattern.Pattern;
import com.starrocks.sql.optimizer.rule.Binder;
import com.starrocks.sql.optimizer.rule.Rule;

import java.util.List;

/**
 * ApplyRuleTask firstly applies a rule, then
 * <p>
 * If the rule is transformation rule and exploreOnly is true:
 * We need to explore (apply logical rules)
 * <p>
 * If the rule is transformation rule and exploreOnly is false:
 * We need to optimize (apply logical & physical rules)
 * <p>
 * If the rule is implementation rule:
 * We directly enforce and cost the physical expression
 */

public class ApplyRuleTask extends OptimizerTask {
    private final GroupExpression groupExpression;
    private final Rule rule;
    private final boolean exploreOnly;

    ApplyRuleTask(TaskContext context, GroupExpression groupExpression,
                  Rule rule, boolean exploreOnly) {
        super(context);
        this.groupExpression = groupExpression;
        this.rule = rule;
        this.exploreOnly = exploreOnly;
    }

    @Override
    public String toString() {
        return "ApplyRuleTask for groupExpression " + groupExpression +
                "\n rule " + rule +
                "\n exploreOnly " + exploreOnly;
    }

    @Override
    public void execute() {
        if (groupExpression.hasRuleExplored(rule) ||
                groupExpression.isUnused()) {
            return;
        }

        // Apply rule and get all new OptExpressions
        Pattern pattern = rule.getPattern();
        Binder binder = new Binder(pattern, groupExpression);
        OptExpression extractExpr = binder.next();
        List<OptExpression> newExpressions = Lists.newArrayList();
        while (extractExpr != null) {
            if (!rule.check(extractExpr, context.getOptimizerContext())) {
                extractExpr = binder.next();
                continue;
            }

            newExpressions.addAll(rule.transform(extractExpr, context.getOptimizerContext()));
            extractExpr = binder.next();
        }

        for (OptExpression expression : newExpressions) {
            // Insert new OptExpression to memo
            Pair<Boolean, GroupExpression> result = context.getOptimizerContext().getMemo().
                    copyIn(groupExpression.getGroup(), expression);

            // The group has been merged
            if (groupExpression.hasEmptyRootGroup()) {
                return;
            }

            // TODO(kks) optimize this
            GroupExpression newGroupExpression = result.second;
            if (newGroupExpression.getOp().isLogical()) {
                // For logic newGroupExpression, optimize it
                pushTask(new OptimizeExpressionTask(context, newGroupExpression, exploreOnly));
                pushTask(new DeriveStatsTask(context, newGroupExpression));
            } else {
                // For physical newGroupExpression, enforce and cost it,
                // Optimize its inputs if needed
                pushTask(new EnforceAndCostTask(context, newGroupExpression));
            }
        }

        groupExpression.setRuleExplored(rule);
    }
}
